/*
PAELLOS - Plateforme d'Apprentissage En Ligne Libre Orientée Sémantique

Copyright (C) 2011 Observatoire de Paris

Ce programme est un logiciel libre ; vous pouvez le redistribuer et/ou le modifier conformément aux dispositions de la Licence Publique Générale GNU, telle que publiée par la Free Software Foundation ; version 2 de la licence, ou encore (à votre choix) toute version ultérieure.

Ce programme est distribué dans l'espoir qu'il sera utile, mais SANS AUCUNE GARANTIE ; sans même la garantie implicite de COMMERCIALISATION ou D'ADAPTATION A UN OBJET PARTICULIER. Pour plus de détail, voir la Licence Publique Générale GNU .

Vous devez avoir reçu un exemplaire de la Licence Publique Générale GNU en même temps que ce programme ; si ce n'est pas le cas, écrivez à la Free Software Foundation Inc., 675 Mass Ave, Cambridge, MA 02139, Etats-Unis.
*/

package paellos;

import java.io.*;
import java.net.*;
import java.security.NoSuchAlgorithmException;
import java.util.*;

import java.text.SimpleDateFormat;

import javax.servlet.*;
import javax.servlet.http.*;

import paellos.base.*;
import paellos.jdb.JDBException;


/**
 * Servlet pour l'inscription aux formations
 * ajoute un utilisateur et ses coordonnées à la base avec valide=0
 * et envoie un email à l'administrateur pour le prévenir
 *
 * @author Damien Guillaume
 */
public class Inscription extends HttpServlet {
    
    static String fichierProperties = "ServletProperties";
    
    // valeurs dans le fichier ServletProperties
    String smtpServer;
    String emailAdmin; // le serveur SMTP doit accepter cet email dans le "From:"
    String emailSecr; // email de la personne entrant les données dans la base
    
    protected String userDBURL;
    protected String userDBUSER;
    protected String userDBPASSWORD;
    protected String[] noms_formations;
    protected String[] titres_formations;
    
    UserDB userDB;
    CoordDB coordDB;
    
    
    public void init(ServletConfig config) throws ServletException {
        super.init(config);
        Properties props = new Properties();
        try {
            props.load(new InputStreamReader(getClass().getResourceAsStream(fichierProperties), "UTF-8"));
            smtpServer = props.getProperty("smtpServer");
            emailAdmin = props.getProperty("emailAdmin");
            emailSecr = props.getProperty("emailSecr");
            userDBURL = props.getProperty("userDBURL");
            userDBUSER = props.getProperty("userDBUSER");
            userDBPASSWORD = props.getProperty("userDBPASSWORD");
            String s_formations = props.getProperty("formations");
            if (s_formations == null || "".equals(s_formations)) {
                noms_formations = null;
                titres_formations = null;
            } else {
                String[] formations = s_formations.split("\\|");
                noms_formations = new String[formations.length];
                titres_formations = new String[formations.length];
                for (int i=0; i<formations.length; i++) {
                    String[] nom_titre = formations[i].split(":");
                    noms_formations[i] = nom_titre[0];
                    if (nom_titre.length > 0)
                        titres_formations[i] = nom_titre[1];
                    else
                        titres_formations[i] = noms_formations[i];
                }
            }
        } catch (IOException ex) {
            erreur(Strings.get("erreur_proprietes") + " ", ex);
        }
        userDB = new UserDB(userDBURL, userDBUSER, userDBPASSWORD);
        try {
            if (!userDB.tableExists()) {
                log(Strings.get("info_creation_table_utilisateurs"));
                userDB.createDB();
            }
        } catch (JDBException ex) {
            erreur(Strings.get("erreur_creation_table_utilisateurs"), ex);
            return;
        }
        coordDB = new CoordDB(userDBURL, userDBUSER, userDBPASSWORD);
        try {
            if (!coordDB.tableExists()) {
                log(Strings.get("info_creation_table_coordonnees"));
                coordDB.createDB();
            }
        } catch (JDBException ex) {
            erreur(Strings.get("erreur_creation_table_coordonnees"), ex);
            return;
        }
    }
    
    protected static String tronquer(String s, final int taille) {
        if (s == null)
            return(null);
        s = s.trim();
        if (s.length() > taille)
            return(s.substring(0, taille));
        return(s);
    }
    
    protected void erreur(final String msg) {
        log(Strings.get("erreur") + " : " + msg);
    }
    
    protected void erreur(String msg, final Exception ex) {
        if (msg == null)
            msg = Strings.get("erreur");
        String nmsg = msg + " : " + ex.getClass().getName() + " : " + ex.getMessage();
        log(nmsg);
    }
    
    /**
     * Paramètres possibles pour une requête POST :
     * - username
     * - motdepasse
     * - motdepasse2
     * - nom
     * - prenom
     * - email
     * - fonction
     * - jourNaissance
     * - moisNaissance
     * - anNaissance
     * - formation
     * - niveau
     * - modules
     * - matiere
     * - classe
     * - centre
     * - rue1
     * - rue2
     * - codepostal
     * - ville
     * - pays
     * - telephone
     * - connection
     * - acces
     */
    public void doPost(HttpServletRequest req, HttpServletResponse resp) throws ServletException, IOException {
        if (req.getCharacterEncoding() == null)
            req.setCharacterEncoding("UTF-8");
        final String username = tronquer(req.getParameter("username"), 15);
        if (username == null || "".equals(username)) {
            envoyerErreur(resp, Strings.get("manque_nom_utilisateur"));
            return;
        }
        if (username.indexOf(' ') != -1) {
            envoyerErreur(resp, Strings.get("espaces_nom_utilisateur"));
            return;
        }
        final String accents = "àâéèêëîïôöùûüçÀÂÉÈÊËÎÏÔÙÛÜÇ";
        for (int i=0; i<accents.length(); i++)
            if (username.indexOf(accents.charAt(i)) != -1) {
                envoyerErreur(resp, Strings.get("accents_nom_utilisateur"));
                return;
            }
        final String majuscules = "ABCDEFGHIJKLMNOPQRSTUVWXYZ";
        for (int i=0; i<majuscules.length(); i++)
            if (username.indexOf(majuscules.charAt(i)) != -1) {
                envoyerErreur(resp, Strings.get("majuscules_nom_utilisateur"));
                return;
            }
        final String autorises = "abcdefghijklmnopqrstuvwxyz0123456789.-_";
        for (int i=0; i<username.length(); i++)
            if (autorises.indexOf(username.charAt(i)) == -1) {
                envoyerErreur(resp, Strings.get("caracteres_nom_utilisateurs"));
                return;
            }
        // BUG: un utilisateur envoyant le même formulaire 2 fois très rapidement peut se retrouver avec 2 comptes ayant le même nom
        User user = new User(username);
        try {
            final Vector<User> fusers = userDB.findUsers(user);
            if (fusers.size() > 0) {
                envoyerErreur(resp, Strings.get("nom_deja_pris"));
                return;
            }
        } catch (JDBException ex) {
            erreur(Strings.get("erreur_verif_nom_utilisateur"), ex);
            return;
        }
        final String motdepasse = tronquer(req.getParameter("motdepasse"), 20);
        if (motdepasse == null || "".equals(motdepasse)) {
            envoyerErreur(resp, Strings.get("manque_mdp"));
            return;
        }
        if (motdepasse.length() < 5) {
            envoyerErreur(resp, Strings.get("mdp_trop_court"));
            return;
        }
        final String motdepasse2 = tronquer(req.getParameter("motdepasse2"), 20);
        if (!motdepasse.equals(motdepasse2)) {
            envoyerErreur(resp, Strings.get("mdp_differents"));
            return;
        }
        if (!"".equals(req.getParameter("codepostal"))) {
            try {
                Integer oi = new Integer(req.getParameter("codepostal"));
            } catch (NumberFormatException ex) {
                envoyerErreur(resp, Strings.get("cp_entier"));
                return;
            }
        }
        boolean premierUtilisateur = false;
        try {
            if (userDB.nbRecords() == 0)
                premierUtilisateur = true;
        } catch (JDBException ex) {
            erreur(Strings.get("erreur_verif_premier_util"), ex);
        }
        initUser(user, req, premierUtilisateur);
        if (user.email == null || user.email.indexOf('@') == -1 || user.email.indexOf('.') == -1) {
            envoyerErreur(resp, Strings.get("mail_invalide"));
            return;
        }
        
        final String jourNaissance = tronquer(req.getParameter("jourNaissance"), 2);
        final String moisNaissance = tronquer(req.getParameter("moisNaissance"), 2);
        final String anNaissance = tronquer(req.getParameter("anNaissance"), 4);
        if (((jourNaissance != null && !"".equals(jourNaissance) && !"01".equals(jourNaissance)) ||
                (moisNaissance != null && !"".equals(moisNaissance) && !"01".equals(moisNaissance))) &&
                (anNaissance == null || anNaissance.length() != 4)) {
            envoyerErreur(resp, Strings.get("date_naissance_invalide"));
            return;
        }
        
        try {
            userDB.addUser(user);
            final Vector<User> fusers = userDB.findUsers(user);
            if (fusers.size() > 0)
                user = fusers.get(0);
            else
                erreur(Strings.get("utilisateur_non_retrouve") + " : " + user);
        } catch (JDBException ex) {
            erreur(Strings.get("erreur_recherche_apres_ajout"), ex);
            return;
        }
        Coord coord = null;
        if (user.numero == null)
            erreur(Strings.get("erreur_ajout_coordonnees"));
        else {
            coord = new Coord(user.numero);
            initCoord(coord, req);
        
            try {
                coordDB.addCoord(coord);
            } catch (JDBException ex) {
                erreur("coordDB.addCoord", ex);
                return;
            }
        }
        
        if (!premierUtilisateur)
            envoyerMailAdmin(user, req);
        
        log(Strings.get("demande_inscription") + " " + user.username);
        
        confirmerDemande(resp, user, coord, motdepasse, premierUtilisateur);
        
        return;
    }
    
    /**
     * Paramètres possibles pour une requête GET :
     * formation: nom de la formation (par défaut la première formation dans ServletProperties)
     * fonction: etudiant|tuteur (etudiant par défaut)
     */
    public void doGet(HttpServletRequest req, HttpServletResponse resp) throws ServletException, IOException {
        String nom_formation = req.getParameter("formation");
        if (nom_formation == null && noms_formations != null)
            nom_formation = noms_formations[0];
        String titre_formation = nom_formation;
        if (noms_formations != null)
            for (int i=0; i<noms_formations.length; i++)
                if (noms_formations[i].equals(nom_formation)) {
                    titre_formation = titres_formations[i];
                    break;
                }
        String fonction = req.getParameter("fonction");
        if (fonction == null)
            fonction = "etudiant";
        if ("etudiant".equals(fonction))
            fonction = "étudiant";
        req.setAttribute("nom_formation", nom_formation);
        req.setAttribute("titre_formation", titre_formation);
        req.setAttribute("fonction", fonction);
        envoyerPageFormulaire(req, resp);
    }
    
    public void envoyerPageFormulaire(final HttpServletRequest req, final HttpServletResponse resp) {
        try {
            getServletContext().getRequestDispatcher("/WEB-INF/jsp/inscription.jsp").forward(req, resp);
        } catch (ServletException ex) {
            log("Administration: envoyerPageConnexion: " + ex.getClass() + ": " + ex.getMessage());
            envoyerErreur(resp, Strings.get("erreur_cf_log"));
        } catch (IOException ex) {
            log("Administration: envoyerPageConnexion: " + ex.getClass() + ": " + ex.getMessage());
            envoyerErreur(resp, Strings.get("erreur_cf_log"));
        }
    }
    
    /**
     * Initialisation des champs d'un utilisateur à partir de la requête.
     * Le premier utilisateur est automatiquement validé comme administrateur.
     */
    protected void initUser(final User user, final HttpServletRequest req, final boolean premierUtilisateur) {
        user.nom = tronquer(req.getParameter("nom"), 30);
        if (user.nom != null && user.nom.length() > 1 && Character.isLowerCase(user.nom.charAt(0)))
            user.nom = user.nom.substring(0, 1).toUpperCase() + user.nom.substring(1);
        user.prenom = tronquer(req.getParameter("prenom"), 20);
        if (user.prenom != null && user.prenom.length() > 1 && Character.isLowerCase(user.prenom.charAt(0)))
            user.prenom = user.prenom.substring(0, 1).toUpperCase() + user.prenom.substring(1);
        final String motdepasse = tronquer(req.getParameter("motdepasse"), 20);
        try {
            user.password = userDB.encryptSHA256(motdepasse);
        } catch (NoSuchAlgorithmException ex) {
            erreur("initUser: NoSuchAlgorithmException", ex);
            user.password = userDB.encrypt(motdepasse);
        }
        user.email = tronquer(req.getParameter("email"), 40);
        if (premierUtilisateur)
            user.fonction = "administrateur";
        else
            user.fonction = tronquer(req.getParameter("fonction"), 20);
        final String jourNaissance = tronquer(req.getParameter("jourNaissance"), 2);
        final String moisNaissance = tronquer(req.getParameter("moisNaissance"), 2);
        final String anNaissance = tronquer(req.getParameter("anNaissance"), 4);
        if ("01".equals(jourNaissance) && "01".equals(moisNaissance) && "19".equals(anNaissance))
            user.naissance = "";
        else
            user.naissance = anNaissance + "-" + moisNaissance + "-" + jourNaissance;
        user.formation = tronquer(req.getParameter("formation"), 15);
        user.niveau = tronquer(req.getParameter("niveau"), 20);
        user.modules = tronquer(req.getParameter("modules"), 20);
        user.matiere = tronquer(req.getParameter("matiere"), 40);
        user.classe = tronquer(req.getParameter("classe"), 40);
        final Date now = new Date();
        final SimpleDateFormat formatter = new SimpleDateFormat("yyyy-MM-dd");
        final String date = formatter.format(now);
        user.dateinscription = date;
        user.centre = tronquer(req.getParameter("centre"), 20);
        if (premierUtilisateur)
            user.valide = "1";
        else
            user.valide = "0";
    }
    
    /**
     * Initialisation des champs de coordonnées d'un utilisateur à partir de la requête.
     */
    protected void initCoord(final Coord coord, final HttpServletRequest req) {
        coord.rue1 = tronquer(req.getParameter("rue1"), 100);
        coord.rue2 = tronquer(req.getParameter("rue2"), 100);
        coord.codepostal = tronquer(req.getParameter("codepostal"), 6);
        if ("".equals(coord.codepostal))
            coord.codepostal = null;
        coord.ville = tronquer(req.getParameter("ville"), 50);
        if (coord.ville != null && coord.ville.length() > 1)
            coord.ville = Character.toUpperCase(coord.ville.charAt(0)) + coord.ville.substring(1);
        coord.pays = tronquer(req.getParameter("pays"), 30);
        if (coord.pays != null && coord.pays.length() > 1)
            coord.pays = Character.toUpperCase(coord.pays.charAt(0)) + coord.pays.substring(1);
        coord.telephone = tronquer(req.getParameter("telephone"), 20);
        coord.connection = tronquer(req.getParameter("connection"), 30);
        coord.acces = tronquer(req.getParameter("acces"), 30);
    }
    
    /**
     * Envoi d'un mail à l'administrateur pour signaler la demande d'inscription.
     */
    protected void envoyerMailAdmin(final User user, final HttpServletRequest req) {
        final String from = emailAdmin;
        final String to = emailSecr;
        final String subject = Strings.get("inscription") + " " + user.formation + " " + user.username;
        String body = user.prenom + " " + user.nom + " " + Strings.get("a_demande_inscription");
        if ("tuteur".equals(user.fonction))
            body += " " + Strings.get("comme_tuteur");
        body += " " + Strings.get("sous_nom_utilisateur") + " '" + user.username + "'";
        body += " " + Strings.get("pour_formation") + " " + user.formation + "\n";
        
        final Mail courriel = new Mail(this, req.getServerName(), smtpServer, 25, from, to, user.email);
        courriel.definirMessage(subject, body);
        try {
            courriel.envoyer(true);
        } catch (IOException ex) {
            erreur("envoyerMailAdmin", ex);
        }
    }
    
    /**
     * Envoi d'une page web confirmant la demande d'inscription.
     */
    public void confirmerDemande(final HttpServletResponse resp, final User user, final Coord coord,
            final String mdp, final boolean premierUtilisateur) {
        try {
            resp.setContentType("text/html; charset=UTF-8");
            final PrintWriter pw = resp.getWriter();
            pw.println("<html>");
            pw.println("<head>");
            pw.println("<meta http-equiv=\"Content-Type\" content=\"text/html; charset=UTF-8\">");
            pw.println("<title>" + Strings.get("titre_page_confirmation") + "</title>");
            pw.println("</head>");
            pw.println("<body>");
            if (premierUtilisateur) {
                pw.println("<p>" + Strings.get("inscription_validee") + "</p>");
                pw.println("<p><a href=\"Administration\">" + Strings.get("titre_administration") + "</a></p>");
            } else {
                pw.println("<p>" + Strings.get("demande_envoyee") + "</p>");
                pw.println("<p>" + Strings.get("votre_mdp") + " '<tt>" + mdp + "</tt>'</p>");
                pw.println("<p>" + Strings.get("notez_le") + "</p>");
            }
            pw.println("<br>");
            pw.println("<pre>");
            
            final String[] userFields = UserDB.fields;
            final Object[] uv = user.getFieldValues();
            for (int i=0; i<userFields.length; i++)
                if (uv[i] != null && !"password".equals(userFields[i]))
                    pw.println(userFields[i] + " :  " + uv[i].toString());
            
            if (coord != null) {
                pw.println("");
                final String[] coordFields = CoordDB.fields;
                final Object[] cv = coord.getFieldValues();
                for (int i=0; i<coordFields.length; i++)
                    if (cv[i] != null)
                        pw.println(coordFields[i] + " :  " + cv[i].toString());
            }
            pw.println("</pre>");
            if (!premierUtilisateur) {
                pw.println("<br>");
                pw.println("<p>" + Strings.get("mail_quand_inscription_validee") + "</p>");
            }
            pw.println("</body>");
            pw.println("</html>");
            pw.flush();
        } catch (IOException ex) {
            erreur("confirmerDemande", ex);
        }
    }
    
    /*
    public String nouveauMotDePasse() {
        String mdp = "";
        Random rnd = new Random();
        // on n'utilise plus 0, 1 et l pour éviter de les confondre avec O, l et I
        String lettres = "23456789abcdefghijkmnopqrstuvwxyz";
        for (int i=0; i<8; i++)
            mdp += lettres.charAt(rnd.nextInt(lettres.length()));
        return(mdp);
    }
    */
    
    public void envoyerErreur(final HttpServletResponse resp, final String errmsg) {
        try {
            resp.setContentType("text/html; charset=UTF-8");
            final PrintWriter pw = resp.getWriter();
            pw.println("<html>");
            pw.println("<head>");
            pw.println("<meta http-equiv=\"Content-Type\" content=\"text/html; charset=UTF-8\">");
            pw.println("<title>" + Strings.get("titre_erreur") + "</title>");
            pw.println("</head>");
            pw.println("<body>");
            pw.println(Strings.get("une_erreur") + " " + errmsg);
            pw.println("<br>");
            pw.println(Strings.get("bouton_retour"));
            pw.println("</body>");
            pw.println("</html>");
            pw.flush();
        } catch (IOException ex) {
            erreur("envoyerErreur", ex);
        }
    }
    
}

