/*
PAELLOS - Plateforme d'Apprentissage En Ligne Libre Orientée Sémantique

Copyright (C) 2011 Observatoire de Paris

Ce programme est un logiciel libre ; vous pouvez le redistribuer et/ou le modifier conformément aux dispositions de la Licence Publique Générale GNU, telle que publiée par la Free Software Foundation ; version 2 de la licence, ou encore (à votre choix) toute version ultérieure.

Ce programme est distribué dans l'espoir qu'il sera utile, mais SANS AUCUNE GARANTIE ; sans même la garantie implicite de COMMERCIALISATION ou D'ADAPTATION A UN OBJET PARTICULIER. Pour plus de détail, voir la Licence Publique Générale GNU .

Vous devez avoir reçu un exemplaire de la Licence Publique Générale GNU en même temps que ce programme ; si ce n'est pas le cas, écrivez à la Free Software Foundation Inc., 675 Mass Ave, Cambridge, MA 02139, Etats-Unis.
*/

package paellos;

import java.io.*;
import java.security.NoSuchAlgorithmException;
import java.util.*;

import java.text.ParseException;
import java.text.SimpleDateFormat;

import javax.servlet.*;
import javax.servlet.http.*;

import paellos.jdb.*;
import paellos.base.*;

/**
 * Servlet pour CELA: pour les tuteurs, liste des lectures des étudiants
 *
 * @author Damien Guillaume
 */
public class Lectures extends HttpServlet {

    // valeurs définies dans le fichier ServletProperties
    protected String cheminSiteTuteurs;
    protected String userDBURL;
    protected String userDBUSER;
    protected String userDBPASSWORD;
    protected String[] noms_formations;
    protected String[] titres_formations;
    
    UserDB userDB;
    LectureDB lectDB;
    ConnectionDB connDB;
    
    public void init(ServletConfig config) throws ServletException {
        super.init(config);
        Properties props = new Properties();
        try {
            props.load(new InputStreamReader(getClass().getResourceAsStream("ServletProperties"), "UTF-8"));
            cheminSiteTuteurs = props.getProperty("cheminSiteTuteurs");
            userDBURL = props.getProperty("userDBURL");
            userDBUSER = props.getProperty("userDBUSER");
            userDBPASSWORD = props.getProperty("userDBPASSWORD");
            String s_formations = props.getProperty("formations");
            if (s_formations == null || "".equals(s_formations)) {
                noms_formations = null;
                titres_formations = null;
            } else {
                String[] formations = s_formations.split("\\|");
                noms_formations = new String[formations.length];
                titres_formations = new String[formations.length];
                for (int i=0; i<formations.length; i++) {
                    String[] nom_titre = formations[i].split(":");
                    noms_formations[i] = nom_titre[0];
                    if (nom_titre.length > 0)
                        titres_formations[i] = nom_titre[1];
                    else
                        titres_formations[i] = noms_formations[i];
                }
            }
        } catch (IOException ex) {
            log(Strings.get("erreur_proprietes") + " " + ex.getMessage());
        }
        userDB = new UserDB(userDBURL, userDBUSER, userDBPASSWORD);
        connDB = new ConnectionDB(userDBURL, userDBUSER, userDBPASSWORD);
        lectDB = new LectureDB(userDBURL, userDBUSER, userDBPASSWORD);
        try {
            if (!connDB.tableExists()) {
                log(Strings.get("info_creation_table_connections"));
                connDB.createDB();
            }
            if (!lectDB.tableExists()) {
                log(Strings.get("info_creation_table_lectures"));
                lectDB.createDB();
            }
        } catch (JDBException ex) {
            log("tableExists/createDB: " + ex.getMessage());
        }
    }
    
    /**
     * Paramètres possibles pour une requête POST :
     * - login : login pour la connexion au site
     * - password : mot de passe pour la connexion au site
     */
    public void doPost(HttpServletRequest req, HttpServletResponse resp) throws ServletException, IOException {
        if (req.getCharacterEncoding() == null)
            req.setCharacterEncoding("UTF-8");
        if (req.getParameter("login") != null) {
            final HttpSession session = req.getSession(true);
            String nom = req.getParameter("login");
            String password = req.getParameter("password");
            if(nom == null || nom.length() == 0)
                nom = "anonymous";
            
            // recherche de l'utilisateur dans la base
            User user = null;
            final User finduser = new User(nom);
            finduser.valide = "1";
            try {
                final Vector<User> v = userDB.findUsers(finduser);
                if (v.size() > 0)
                    user = v.get(0); // on prend le premier utilisateur avec le username
            } catch (JDBException ex) {
                log("JDBException: " + ex.getMessage());
            }
            if (user == null) {
                log(Strings.get("utilisateur_non_trouve") + " " + nom + " -> " + Strings.get("acces_refuse"));
                envoyerErreur(resp, Strings.get("acces_refuse"));
                return;
            }
            
            // test mot de passe
            try {
                if (password == null || !JDB.testEnc(password, user.password)) {
                    log(Strings.get("mauvais_mot_de_passe") + " " + nom);
                    envoyerErreur(resp, Strings.get("acces_refuse"));
                    return;
                }
            } catch (NoSuchAlgorithmException ex) {
                log("Lectures.doPost: NoSuchAlgorithmException: " + ex.getMessage());
                envoyerErreur(resp, Strings.get("acces_refuse"));
                return;
            }
            
            // vérif si utilisateur tuteur ou système ou supertuteur
            if (user.fonction == null || !(user.fonction.equals("tuteur") || user.fonction.equals("supertuteur") ||
                    user.fonction.equals("administrateur") || user.fonction.equals("système"))) {
                log(Strings.get("utilisateur_non_tuteur") + " " + nom);
                envoyerErreur(resp, Strings.get("acces_refuse"));
                return;
            }
            
            // obtention date d'aujourd'hui
            final Date now = new Date();
            final SimpleDateFormat formatter = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm");
            final String date = formatter.format(now);
            
            // test fermeture de compte
            if (user.datefermeture != null && !"".equals(user.datefermeture)) {
                Date fermeture;
                try {
                    fermeture = formatter.parse(user.datefermeture);
                    if (now.after(fermeture)) {
                        log(Strings.get("utilisation_compte_ferme") + " " + nom);
                        envoyerErreur(resp, Strings.get("acces_refuse_compte_ferme"));
                        return;
                    }
                } catch (ParseException ex) {
                    log("datefermeture ParseException: " + ex.getMessage());
                }
            }
            
            session.setAttribute("user", user);
            log(Strings.get("lectures_login") + " " + user.username);
            envoyerPageEtudiants(resp, user, null);
            return;
            
        }
        envoyerErreur(resp, Strings.get("requete_non_comprise"));
    }
    
    /**
     * Paramètres possibles pour une requête GET (un seul paramètre entre numtuteur et numetudiant est nécessaire) :
     * numtuteur : numéro du tuteur (doit correspondre à l'utilisateur)
     * numetudiant : numéro de l'étudiant dont on veut voir les lectures
     * tri : nom|date (nom par défaut)
     */
    public void doGet(HttpServletRequest req, HttpServletResponse resp) throws ServletException, IOException {
        final HttpSession session = req.getSession(false);
        if (session == null) {
            if (req.getQueryString() == null) {
                envoyerPageConnexion(req, resp);
                return;
            }
            log(Strings.get("erreur_pas_de_session") + " " + req.getRequestURL());
            envoyerErreur(resp, Strings.get("session_expiree"));
            return;
        }
        final User tuteur = (User)session.getAttribute("user");
        if (tuteur == null) {
            if (req.getQueryString() == null) {
                envoyerPageConnexion(req, resp);
                return;
            }
            log(Strings.get("erreur_user_null") + " " + req.getRequestURL());
            envoyerErreur(resp, Strings.get("erreur_donnees_session"));
            return;
        }
        if ("étudiant".equals(tuteur.fonction)) {
            log(Strings.get("erreur_etudiant_tente_acces") + " " + tuteur.username);
            envoyerErreur(resp, Strings.get("acces_interdit"));
            return;
        }
        String numetudiant = req.getParameter("numetudiant");
        String numtuteur = req.getParameter("numtuteur");
        String tri = req.getParameter("tri");
        if (numtuteur == null)
            numtuteur = tuteur.numero;
        if (numetudiant != null) {
            User etudiant = chercherUser(numetudiant);
            if (etudiant == null) {
                log(Strings.get("erreur_mauvais_no_etudiant") + ": " + numetudiant);
                envoyerErreur(resp, Strings.get("erreur_mauvais_no_etudiant"));
                return;
            }
            envoyerPageLectures(resp, etudiant, tuteur, tri);
            return;
        } else if (numtuteur != null) {
            if (!tuteur.numero.equals(numtuteur)) {
                log(Strings.get("erreur_mauvais_no_tuteur"));
                envoyerErreur(resp, Strings.get("erreur_mauvais_no_tuteur"));
            } else
                envoyerPageEtudiants(resp, tuteur, tri);
            return;
        }
        envoyerErreur(resp, Strings.get("requete_non_comprise"));
    }
    
    public User chercherUser(final String numuser) {
        final User finduser = new User();
        finduser.numero = numuser;
        finduser.valide = "1";
        try {
            final Vector<User> v = userDB.findUsers(finduser);
            if (v.size() != 0) {
                User tuteur = v.get(0);
                return(tuteur);
            }
        } catch (JDBException ex) {
            log("chercherUser: JDBException: " + ex.getMessage());
        }
        log("chercherUser: pas d'utilisateur " + numuser);
        return(null);
    }
    
    public Vector<Connection> chercherConnections(final String numuser) {
        final Connection findconn = new Connection(null, numuser, null, null);
        try {
            return(connDB.findConnections(findconn));
        } catch (JDBException ex) {
            log("chercherConnections: JDBException: " + ex.getMessage());
            return(null);
        }
    }
    
    /*
    public Vector<Lecture> chercherLectures(final String numconnection) {
        final Lecture findlect = new Lecture(null, null, numconnection);
        try {
            return(lectDB.findLectures(findlect));
        } catch (JDBException ex) {
            log("chercherConnections: JDBException: " + ex.getMessage());
            return(null);
        }
    }
    */
    
    public void envoyerErreur(final HttpServletResponse resp, final String errmsg) {
        try {
            resp.setHeader("Pragma", "no-cache");
            resp.setHeader("Cache-Control", "no-cache");
            resp.setHeader("Expires", "0");
            resp.setContentType("text/html; charset=UTF-8");
            
            final PrintWriter pw = resp.getWriter();
            pw.println("<html>");
            pw.println("<head>");
            pw.println("<meta http-equiv=\"Content-Type\" content=\"text/html; charset=UTF-8\">");
            pw.println("<title>" + Strings.get("titre_erreur") + "</title>");
            pw.println(" <link href=\"paellos.css\" rel=\"stylesheet\" type=\"text/css\">");
            pw.println("</head>");
            pw.println("<body>");
            pw.println(Strings.get("une_erreur") + " " + errmsg);
            pw.println("<br>");
            pw.println(Strings.get("bouton_retour"));
            pw.println("</body>");
            pw.println("</html>");
            pw.flush();
        } catch (IOException ex) {
            log("envoyerErreur: " + ex.getMessage());
        }
    }
    
    private Date dateDerniereConnexion(final String numuser) {
        final SimpleDateFormat isoformatter = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm");
        final Vector<Connection> connections = chercherConnections(numuser);
        Date derniere = null;
        for (Connection conn : connections) {
            Date debut = null;
            if (conn.debut != null) {
                try {
                    debut = isoformatter.parse(conn.debut);
                } catch (ParseException ex) {
                }
            }
            if (debut != null)
                if (derniere == null || debut.after(derniere))
                    derniere = debut;
        }
        return(derniere);
    }
    
    public void envoyerPageConnexion(final HttpServletRequest req, final HttpServletResponse resp) {
        try {
            getServletContext().getRequestDispatcher("/WEB-INF/jsp/lectures.jsp").forward(req, resp);
        } catch (ServletException ex) {
            log("Lectures: envoyerPageConnexion: " + ex.getClass() + ": " + ex.getMessage());
            envoyerErreur(resp, Strings.get("erreur_cf_log"));
        } catch (IOException ex) {
            log("Lectures: envoyerPageConnexion: " + ex.getClass() + ": " + ex.getMessage());
            envoyerErreur(resp, Strings.get("erreur_cf_log"));
        }
    }
    
    public void envoyerPageEtudiants(final HttpServletResponse resp, final User tuteur, final String tri) {
        try {
            resp.setContentType("text/html; charset=UTF-8");
            final Writer out = resp.getWriter();
            final PrintWriter pw = new PrintWriter(out);
            pw.println("<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01 Transitional//EN\">");
            pw.println("<html>");
            pw.println("<head>");
            pw.println(" <meta http-equiv=\"Content-Type\" content=\"text/html; charset=UTF-8\">");
            pw.println(" <title>" + Strings.get("liste_etudiants") + "</title>");
            pw.println(" <link href=\"paellos.css\" rel=\"stylesheet\" type=\"text/css\">");
            pw.println("</head>");
            pw.println("<body>");
            if ("tuteur".equals(tuteur.fonction))
                pw.println("<h2>" + Strings.get("etudiants_de") + " " + tuteur.prenom + " " + tuteur.nom + "</h2>");
            else
                pw.println("<h2>" + Strings.get("titre_etudiants") + "</h2>");
            pw.println("<table border=\"1\">");
            pw.println(" <tr><th><a href=\"Lectures?tri=nom\">" + Strings.get("titre_nom") +
                "</a></th><th><a href=\"Lectures?tri=date\">" + Strings.get("derniere_connexion") + "</a></th></tr>");
            final User finduser = new User();
            finduser.fonction = "étudiant";
            if ("tuteur".equals(tuteur.fonction))
                finduser.numtuteur = tuteur.numero;
            // cas "supertuteur", "administrateur" et "système" : tous les étudiants
            finduser.valide = "1";
            if (noms_formations != null && noms_formations.length == 1)
                finduser.formation = noms_formations[0];
            final SimpleDateFormat frenchformatter = new SimpleDateFormat(Strings.get("format_date"));
            final SimpleDateFormat isoformatter = new SimpleDateFormat("yyyy-MM-dd");
            final String maintenant = isoformatter.format(new Date());
            try {
                final Vector<User> etudiants = userDB.findUsers(finduser);
                // sélection des étudiants avec la bonne formation et un compte ouvert
                for (int i=0; i<etudiants.size(); i++) {
                    User etudiant = etudiants.get(i);
                    boolean formationok;
                    if (noms_formations == null || etudiant.formation == null)
                        formationok = true;
                    else
                        formationok = Arrays.asList(noms_formations).contains(etudiant.formation);
                    boolean dateok = (etudiant.datefermeture == null || "".equals(etudiant.datefermeture) ||
                        etudiant.datefermeture.compareTo(maintenant) > 0);
                    if (!formationok || !dateok) {
                        etudiants.remove(i);
                        i--;
                    }
                }
                // recherche des dates de dernière connexion correspondantes
                final Vector<Date> dates = new Vector<Date>();
                for (User etudiant : etudiants)
                    dates.add(dateDerniereConnexion(etudiant.numero));
                
                // tri par nom (par défaut) ou par date
                if ("date".equals(tri)) {
                    for (int i=0; i<dates.size(); i++)
                        for (int j=i+1; j<dates.size(); j++) {
                            Date datei = dates.get(i);
                            Date datej = dates.get(j);
                            if ((datei == null && datej != null) || ((datei != null && datej != null) && datei.compareTo(datej) < 0)) {
                                final User tempUser = etudiants.get(i);
                                etudiants.setElementAt(etudiants.get(j), i);
                                etudiants.setElementAt(tempUser, j);
                                final Date tempDate = dates.get(i);
                                dates.setElementAt(dates.get(j), i);
                                dates.setElementAt(tempDate, j);
                            }
                        }
                } else {
                    for (int i=0; i<etudiants.size(); i++)
                        for (int j=i+1; j<etudiants.size(); j++)
                            if ((etudiants.get(i)).compareTo(etudiants.get(j)) > 0) {
                                final User tempUser = etudiants.get(i);
                                etudiants.setElementAt(etudiants.get(j), i);
                                etudiants.setElementAt(tempUser, j);
                                final Date tempDate = dates.get(i);
                                dates.setElementAt(dates.get(j), i);
                                dates.setElementAt(tempDate, j);
                            }
                }
                // affichage
                for (int i=0; i<etudiants.size(); i++) {
                    final User etudiant = etudiants.get(i);
                    int ntab = (i%2==0 ? 1 : 2);
                    pw.println(" <tr>");
                    pw.println("  <td class=\"t" + ntab + "\"><a href=\"Lectures?numetudiant="+etudiant.numero+"\">" +
                        etudiant.prenom + " " + etudiant.nom + "</a></td>");
                    final Date derniere = dates.get(i);
                    String sdate;
                    if (derniere == null)
                        sdate = Strings.get("interrogation");
                    else
                        sdate = frenchformatter.format(derniere);
                    pw.println("  <td class=\"t" + ntab + "\">" + sdate + "</td>");
                    pw.println(" </tr>");
                }
            } catch (JDBException ex) {
                log("envoyerPageEtudiants: JDBException: " + ex.getMessage());
            }
            pw.println("</table>");
            pw.println("<p><a href=\"" + cheminSiteTuteurs + "index.html\">" + Strings.get("revenir_accueil") + "</a></p>");
            pw.println("</body>");
            pw.println("</html>");
            pw.flush();
        } catch (IOException ex) {
            log("envoyerPageEtudiants: IOException: " + ex.getMessage());
        }
    }
    
    public void envoyerPageLectures(final HttpServletResponse resp, final User etudiant, final User tuteur, final String tri) {
        try {
            resp.setContentType("text/html; charset=UTF-8");
            final Writer out = resp.getWriter();
            final PrintWriter pw = new PrintWriter(out);
            pw.println("<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01 Transitional//EN\">");
            pw.println("<html>");
            pw.println("<head>");
            pw.println(" <meta http-equiv=\"Content-Type\" content=\"text/html; charset=UTF-8\">");
            pw.println(" <title>" + Strings.get("lectures_etudiant") + "</title>");
            pw.println(" <link href=\"paellos.css\" rel=\"stylesheet\" type=\"text/css\">");
            pw.println("</head>");
            pw.println("<body>");
            pw.println("<h2>" + Strings.get("lectures_de") + " " + etudiant.prenom + " " + etudiant.nom + "</h2>");
            pw.println("<table border=\"1\">");
            pw.println(" <tr><th><a href=\"Lectures?numetudiant="+etudiant.numero+"&amp;tri=page\">" + Strings.get("titre_page") + "</a></th>" +
                "<th><a href=\"Lectures?numetudiant="+etudiant.numero+"&amp;tri=date\">" + Strings.get("derniere_lecture") + "</a></th></tr>");
            final Vector<String> resalu = new Vector<String>();
            final Vector<String> resdate = new Vector<String>();
            final SimpleDateFormat isoformatter = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm");
            final SimpleDateFormat frenchformatter = new SimpleDateFormat(Strings.get("format_date"));
            /*
            // recherche des lectures : pas optimisé (pb du grand nombre de requêtes SQL)
            Vector<Connection> connections = chercherConnections(etudiant.numero);
            if (connections != null) {
                for (int ic=0; ic<connections.size(); ic++) {
                    Connection conn = connections.get(ic);
                    Vector lectures = chercherLectures(conn.numero);
                    if (lectures != null) {
                        for (int il=0; il<lectures.size(); il++) {
                            Lecture lect = lectures.get(il);
                            String alu = lect.alu;
                            if (alu.endsWith("/index"))
                                alu = alu.substring(0, alu.indexOf("/index"));
                            int indalu = resalu.indexOf(alu);
                            if (indalu != -1) {
                                try {
                                    Date ddejalu = isoformatter.parse(resdate.get(indalu));
                                    Date dalu = isoformatter.parse(conn.debut);
                                    if (dalu.after(ddejalu)) {
                                        resalu.setElementAt(alu, indalu);
                                        resdate.setElementAt(conn.debut, indalu);
                                    }
                                } catch (ParseException ex) {
                                }
                            } else {
                                resalu.addElement(alu);
                                resdate.addElement(conn.debut);
                            }
                        }
                    }
                }
            }
            */
            try {
                String[][] tdl = lectDB.dernieresLectures(etudiant.numero); // nouvelle méthode de LectureDB, avec requête SQL spéciale
                for (int i=0; i<tdl.length; i++) {
                    String alu = tdl[i][0];
                    if (alu.endsWith("/index"))
                        alu = alu.substring(0, alu.lastIndexOf("/index"));
                    resalu.add(alu);
                    resdate.add(tdl[i][1]);
                }
            } catch (JDBException ex) {
                log("envoyerPageLectures: JDBException: " + ex.getMessage());
            }
            // tri des lectures
            if ("date".equals(tri)) {
                for (int i=0; i<resdate.size(); i++)
                    for (int j=i+1; j<resdate.size(); j++)
                        if ((resdate.get(i)).compareTo(resdate.get(j)) < 0) {
                            String temp = resalu.get(i);
                            resalu.setElementAt(resalu.get(j), i);
                            resalu.setElementAt(temp, j);
                            temp = resdate.get(i);
                            resdate.setElementAt(resdate.get(j), i);
                            resdate.setElementAt(temp, j);
                        }
            } else {
                for (int i=0; i<resalu.size(); i++)
                    for (int j=i+1; j<resalu.size(); j++)
                        if ((resalu.get(i)).compareTo(resalu.get(j)) > 0) {
                            String temp = resalu.get(i);
                            resalu.setElementAt(resalu.get(j), i);
                            resalu.setElementAt(temp, j);
                            temp = resdate.get(i);
                            resdate.setElementAt(resdate.get(j), i);
                            resdate.setElementAt(temp, j);
                        }
            }
            for (int i=0; i<resalu.size(); i++) {
                int ntab = (i%2==0 ? 1 : 2);
                pw.println(" <tr>");
                final String alu = resalu.get(i);
                pw.println("  <td class=\"t" + ntab + "\">" + alu + "</td>");
                if (resdate.get(i) == null)
                    pw.println("  <td class=\"t" + ntab + "\"> - </td>");
                else {
                    try {
                        final Date dalu = isoformatter.parse(resdate.get(i));
                        pw.println("  <td class=\"t" + ntab + "\">" + frenchformatter.format(dalu) + "</td>");
                    } catch (ParseException ex) {
                        pw.println("  <td class=\"t" + ntab + "\">" + resdate.get(i) + "</td>");
                    }
                }
                pw.println(" </tr>");
            }
            pw.println("</table>");
            pw.println("<p><a href=\"Lectures?numtuteur="+tuteur.numero+"\">" + Strings.get("revenir_liste") + "</a></p>");
            pw.println("</body>");
            pw.println("</html>");
            pw.flush();
        } catch (IOException ex) {
            log("envoyerPageLectures: " + ex.getMessage());
        }
    }
    
}

