/*
PAELLOS - Plateforme d'Apprentissage En Ligne Libre Orientée Sémantique

Copyright (C) 2011 Observatoire de Paris

Ce programme est un logiciel libre ; vous pouvez le redistribuer et/ou le modifier conformément aux dispositions de la Licence Publique Générale GNU, telle que publiée par la Free Software Foundation ; version 2 de la licence, ou encore (à votre choix) toute version ultérieure.

Ce programme est distribué dans l'espoir qu'il sera utile, mais SANS AUCUNE GARANTIE ; sans même la garantie implicite de COMMERCIALISATION ou D'ADAPTATION A UN OBJET PARTICULIER. Pour plus de détail, voir la Licence Publique Générale GNU .

Vous devez avoir reçu un exemplaire de la Licence Publique Générale GNU en même temps que ce programme ; si ce n'est pas le cas, écrivez à la Free Software Foundation Inc., 675 Mass Ave, Cambridge, MA 02139, Etats-Unis.
*/

package paellos;

import java.io.*;
import java.net.*;
import java.util.ArrayList;

import javax.servlet.http.HttpServlet;

/**
 * Envoi d'un mail via un serveur SMTP.
 * Actuellement on ne peut mettre qu'un seul fichier attaché (avec la méthode ajouterFichierAttache)
 * (sauf dans du HTML, avec la méthode ajouterFichierAttacheHTML).
 *
 * @author Damien Guillaume
 */
public class Mail extends Thread {

    private HttpServlet servlet; // la servlet qui fait l'envoie du mail (pour les logs)
    private String machine; // nom du serveur qui lance ce programme
    private String serveurSMTP; // nom du serveur SMTP
    private int port; // port à utiliser sur le serveur SMTP
    private String envoyeur; // adresse mail de l'envoyeur (champ from)
    private String destinataire; // adresse mail du destinaraire (champ to)
    private String repondreA; // adresse mail à laquelle répondre (champ reply-to)
    private String sujet; // sujet du message (champ subject)
    private String message; // corps du mail à envoyer, sous forme de texte simple
    private boolean html; // si le message est en HTML
    private String texte; // dans le cas d'un message HTML, on peut définir une alternative en texte simple
    private FichierAttache fAttache; // fichier attaché au mail (hors HTML)
    private ArrayList<FichierAttache> listeAttachesHTML; // liste des fichiers attachés au HTML
    final private static String separateur1 = "AaB03x"; // séparateur quand on utilise du HTML avec texte alternatif
    final private static String separateur2 = "AaB03y"; // séparateur quand on ajoute une image attachée au mail
    final private static String separateur3 = "AaB03z"; // pour les images associées au HTML
    final private static String encodageMail = "UTF-8"; // (doit être sur 8 bits)
    
    public Mail(HttpServlet servlet, String machine, String serveurSMTP, int port, String envoyeur, String destinataire, String repondreA) {
        this.servlet = servlet;
        this.machine = machine;
        this.serveurSMTP = serveurSMTP;
        this.port = port;
        this.envoyeur = envoyeur;
        this.destinataire = destinataire;
        this.repondreA = repondreA;
        this.sujet = null;
        this.message = null;
        this.html = false;
        this.texte = null;
        this.fAttache = null;
        this.listeAttachesHTML = null;
    }
    
    public void definirMessage(final String sujet, final String message) {
        definirMessage(sujet, message, false);
    }
    
    public void definirMessage(final String sujet, final String message, final boolean html) {
        if (sujet == null)
            this.sujet = "";
        else
            this.sujet = sujet;
        if (message == null)
            this.message = "";
        else
            this.message = message;
        this.html = html;
    }
    
    /**
     * Dans le cas d'un message HTML, on peut définir une alternative en texte simple
     */
    public void definirTexteAlternatif(final String texte) {
        this.texte = texte;
    }
    
    /**
     * Ajoute un fichier attaché au mail (il ne peut actuellement y en avoir qu'un seul)
     */
    public void ajouterFichierAttache(final String nom, final InputStream contenu, final String mime) {
        fAttache = new FichierAttache(null, nom, contenu, mime);
    }
    
    /**
     * Ajoute un fichier attaché au contenu HTML
     */
    public void ajouterFichierAttacheHTML(final String cid, final String nom, final InputStream contenu, final String mime) {
        if (listeAttachesHTML == null)
            listeAttachesHTML = new ArrayList<FichierAttache>();
        listeAttachesHTML.add(new FichierAttache(cid, nom, contenu, mime));
    }
    
    /**
     * Lance l'envoi du mail, avec un thread si l'envoi est asynchrone.
     * Peut lancer une exception en cas d'erreur si l'envoi est synchrone.
     */
    public void envoyer(final boolean asynchrone) throws IOException {
        if (asynchrone)
            start();
        else
            envoi(false);
    }
    
    public void run() {
        try {
            envoi(true);
        } catch (IOException ex) {
            logErreur("Mail: Erreur: " + ex.getMessage());
        }
    }
    
    private void envoi(final boolean asynchrone) throws IOException {
        Socket socket = null;
        try {
            socket = new Socket(InetAddress.getByName(serveurSMTP), port);
            socket.setSoTimeout(30000);
            
            logErreur("Mail: connection à " + serveurSMTP);
            final BufferedWriter to = new BufferedWriter(new OutputStreamWriter(socket.getOutputStream(), encodageMail));
            final BufferedReader from = new BufferedReader(new InputStreamReader(socket.getInputStream()));
            
            String commande = "connexion";
            String reponse = lireReponse(from, commande);
            String code = codeReponse(reponse);
            if (!"220".equals(code))
                erreurReponse(commande, code, reponse);
            
            commande = "helo " + machine;
            envoyerCommande(commande, to);
            reponse = lireReponse(from, commande);
            code = codeReponse(reponse);
            if (!"250".equals(code))
                erreurReponse(commande, code, reponse);
            
            commande = "mail from: <" + envoyeur + ">";
            envoyerCommande(commande, to);
            reponse = lireReponse(from, commande);
            code = codeReponse(reponse);
            if (!"250".equals(code))
                erreurReponse(commande, code, reponse);
            
            commande = "rcpt to: <" + destinataire + ">";
            envoyerCommande(commande, to);
            reponse = lireReponse(from, commande);
            code = codeReponse(reponse);
            if (!"250".equals(code) && !"251".equals(code))
                erreurReponse(commande, code, reponse);
            
            commande = "data";
            envoyerCommande(commande, to);
            reponse = lireReponse(from, commande);
            code = codeReponse(reponse);
            if (!"354".equals(code))
                erreurReponse(commande, code, reponse);
            
            envoyerLigne("MIME-Version: 1.0", to);
            envoyerLigne("To: " + destinataire, to);
            if (repondreA != null)
                envoyerLigne("Reply-To: " + repondreA, to);
            envoyerLigne("Subject: " + encode1522(sujet), to);
            
            envoyerContenu(html, message, texte, fAttache, listeAttachesHTML, to);
            
            envoyerLigne(".", to);
            
            commande = "data (fin)";
            reponse = lireReponse(from, commande);
            code = codeReponse(reponse);
            if (!"250".equals(code))
                erreurReponse(commande, code, reponse);
            
            commande = "quit";
            envoyerCommande(commande, to);
            reponse = lireReponse(from, commande);
            code = codeReponse(reponse);
            if (!"221".equals(code))
                erreurReponse(commande, code, reponse);
            
            socket.close();
            logErreur("Mail: message bien envoyé: from " + envoyeur + " to " + destinataire);
        } catch (IOException ex) {
            if (socket != null) {
                try {
                    socket.close();
                } catch (IOException e) {
                }
            }
            throw ex;
        }
    }
    
    private static void envoyerContenu(final boolean html, final String message, final String texte, final FichierAttache fAtttache,
            final ArrayList<FichierAttache> listeAttachesHTML, final BufferedWriter bw) throws IOException {
        if (html) {
            if (texte != null) {
                envoyerLigne("Content-type: multipart/alternative; boundary=" + separateur1, bw);
                envoyerLigne("", bw);
                envoyerLigne("--" + separateur1, bw);
                envoyerTexte(texte, bw);
                envoyerLigne("--" + separateur1, bw);
                if (fAtttache != null)
                    envoyerHTMLEtAttache(message, fAtttache, listeAttachesHTML, bw);
                else
                    envoyerHTML(message, listeAttachesHTML, bw);
                envoyerLigne("--" + separateur1 + "--", bw);
                envoyerLigne("", bw);
            } else
                envoyerHTMLEtAttache(message, fAtttache, listeAttachesHTML, bw);
        } else {
            if (fAtttache != null)
                envoyerTexteEtAttache(message, fAtttache, bw);
            else
                envoyerTexte(message, bw);
        }
    }
    
    private static void envoyerHTMLEtAttache(final String message, final FichierAttache fAttache,
            final ArrayList<FichierAttache> listeAttachesHTML, final BufferedWriter bw) throws IOException {
        envoyerLigne("Content-type: multipart/mixed; boundary=" + separateur2, bw);
        envoyerLigne("", bw);
        envoyerLigne("--" + separateur2, bw);
        envoyerHTML(message, listeAttachesHTML, bw);
        envoyerLigne("--" + separateur2, bw);
        fAttache.envoyer(bw);
        envoyerLigne("--" + separateur2 + "--", bw);
    }
    
    private static void envoyerTexteEtAttache(final String message, final FichierAttache fAttache,
            final BufferedWriter bw) throws IOException {
        envoyerLigne("Content-type: multipart/mixed; boundary=" + separateur2, bw);
        envoyerLigne("", bw);
        envoyerLigne("--" + separateur2, bw);
        envoyerTexte(message, bw);
        envoyerLigne("", bw);
        envoyerLigne("--" + separateur2, bw);
        fAttache.envoyer(bw);
        envoyerLigne("--" + separateur2 + "--", bw);
    }
    
    private static void envoyerTexte(final String texte, final BufferedWriter bw) throws IOException {
        envoyerLigne("Content-type: text/plain;", bw);
        envoyerLigne("\tcharset=" + encodageMail, bw);
        envoyerLigne("Content-Transfer-Encoding: 8bit", bw);
        envoyerLigne("", bw);
        envoyerLigne(transformerMessage(texte), bw);
    }
    
    private static void envoyerHTML(final String message, final ArrayList<FichierAttache> listeAttaches,
            final BufferedWriter bw) throws IOException {
        if (listeAttaches == null || listeAttaches.size() == 0) {
            envoyerLigne("Content-type: text/html;", bw);
            envoyerLigne("\tcharset=" + encodageMail, bw);
            envoyerLigne("Content-Transfer-Encoding: 8bit", bw);
            envoyerLigne("", bw);
            envoyerLigne(transformerMessage(message), bw);
        } else {
            envoyerLigne("Content-type: multipart/related;", bw);
            envoyerLigne("\ttype=\"text/html\";", bw);
            envoyerLigne("\tboundary=" + separateur3, bw);
            envoyerLigne("", bw);
            envoyerLigne("--" + separateur3, bw);
            envoyerLigne("Content-type: text/html;", bw);
            envoyerLigne("\tcharset=" + encodageMail, bw);
            envoyerLigne("Content-Transfer-Encoding: 8bit", bw);
            envoyerLigne("", bw);
            envoyerLigne(transformerMessage(message), bw);
            envoyerLigne("--" + separateur3, bw);
            for (int i=0; i<listeAttaches.size(); i++) {
                FichierAttache f = listeAttaches.get(i);
                f.envoyer(bw);
                if (i < listeAttaches.size() - 1)
                    envoyerLigne("--" + separateur3, bw);
                else
                    envoyerLigne("--" + separateur3 + "--", bw);
            }
            envoyerLigne("", bw);
        }
    }
    
    private void logErreur(final String message) {
        if (servlet != null)
            servlet.log(message);
        else
            System.err.println(message);
    }
    
    private static void envoyerCommande(final String commande, final BufferedWriter bw) throws IOException {
        bw.write(commande + "\r\n");
        bw.flush();
    }
    
    private static void envoyerLigne(final String ligne, final BufferedWriter bw) throws IOException {
        bw.write(ligne + "\r\n");
        bw.flush();
    }
    
    private static String lireReponse(final BufferedReader br, final String commande) throws IOException {
        String ligne;
        final StringBuffer reponse = new StringBuffer();
        boolean continuer = true;
        
        while (continuer) {
            try {
                ligne = br.readLine();
            } catch (IOException ex) {
                throw new IOException("Erreur I/O à la lecture de la réponse pour la commande " + commande + ": " +
                    ex.getClass().getName() + ": " + ex.getMessage());
            }
            if (ligne == null || ligne.length() < 4)
                throw new IOException("Mauvaise réponse du serveur à la commande " + commande + ": " + ligne);
            continuer = (ligne.charAt(3) == '-');
            if (reponse.length() > 0)
                reponse.append('\n');
            reponse.append(ligne);
        }
        
        return(reponse.toString());
    }
    
    private static String codeReponse(final String reponse) {
        if (reponse == null)
            return(null);
        final int ind = reponse.lastIndexOf('\n');
        String derniereLigne;
        if (ind == -1)
            derniereLigne = reponse;
        else
            derniereLigne = reponse.substring(ind + 1);
        if (derniereLigne.length() > 3)
            return(derniereLigne.substring(0, 3));
        else
            return(null);
    }
    
    private static void erreurReponse(final String commande, final String code, final String reponse) throws IOException {
        String premiereLigne;
        int ind = reponse.indexOf('\n');
        if (ind == -1)
            premiereLigne = reponse;
        else
            premiereLigne = reponse.substring(0, ind);
        throw new IOException("Code d'erreur du serveur SMTP pour la commande " + commande + ": " + code + "; reponse: " + premiereLigne);
    }
    
    private static String transformerMessage(final String message) {
        // pas optimisé :-)
        if (message == null)
            return(null);
        String msg = message.replace("\r\n", "\r");
        msg = msg.replace("\n", "\r");
        msg = msg.replace("\r", "\r\n");
        msg = msg.replace("\r\n.\r\n", "\r\n..\r\n");
        // pb des lignes trop longues : on les coupe au niveau du dernier espace
        final int maximum = 950; // en théorie 998 mais en pratique moins que ça
        int ind1 = 0;
        int ind = 0;
        while (ind != -1) {
            ind = msg.indexOf("\r\n", ind1);
            if (ind != -1) {
                final int taille = ind - ind1;
                if (taille > maximum) {
                    final int inds = msg.lastIndexOf(" ", ind1 + maximum);
                    if (inds != -1 && inds > ind1) {
                        msg = msg.substring(0, inds) + "\r\n" + msg.substring(inds + 1);
                        ind++;
                    } else
                        System.err.println("Mail: erreur: ligne trop longue sans espace");
                }
                ind1 = ind + 2;
            }
        }
        return(msg);
    }
    
    private static void encode1522(final StringBuffer sb, final int b) {
        sb.append('=');
        char hex1 = Character.toUpperCase(Character.forDigit((b >> 4) & 0xF, 16));
        char hex2 = Character.toUpperCase(Character.forDigit(b & 0xF, 16));
        sb.append(hex1);
        sb.append(hex2);
    }
    
    // implémentation limitée de RFC 1522, pour ISO-8859-1
    private static String encode1522(final String s) {
        final String encodage = "ISO-8859-1";
        StringBuffer sb = new StringBuffer();
        sb.append("=?");
        sb.append(encodage);
        sb.append("?Q?");
        byte[] bytes;
        try {
            bytes = s.getBytes(encodage);
        } catch (UnsupportedEncodingException ex) {
            System.err.println("UnsupportedEncodingException: " + ex.getMessage());
            return("");
        }
        for (int i = 0; i < bytes.length; i++) {
            int b = bytes[i];
            if (b < 0)
                b = 256 + b;
            if ((b >= 0x41 && b<= 0x5A) || (b >= 0x61 && b<= 0x7A))
                sb.append((char)b);
            else if (b == 0x20)
                sb.append('_');
            else
                encode1522(sb, b);
        }
        sb.append("?=");
        return(sb.toString());
    }
    
    private static void encode2184(final StringBuffer sb, final int b) {
        sb.append('%');
        char hex1 = Character.toUpperCase(Character.forDigit((b >> 4) & 0xF, 16));
        char hex2 = Character.toUpperCase(Character.forDigit(b & 0xF, 16));
        sb.append(hex1);
        sb.append(hex2);
    }
    
    // implémentation limitée de RFC 2184, pour ISO-8859-1 fr-fr
    private static String encode2184(final String s) {
        final String encodage = "ISO-8859-1";
        StringBuffer sb = new StringBuffer();
        sb.append(encodage);
        sb.append("'fr-fr'");
        byte[] bytes;
        try {
            bytes = s.getBytes(encodage);
        } catch (UnsupportedEncodingException ex) {
            System.err.println("UnsupportedEncodingException: " + ex.getMessage());
            return("");
        }
        for (int i = 0; i < bytes.length; i++) {
            int b = bytes[i];
            if (b < 0)
                b = 256 + b;
            if ((b >= 0x41 && b<= 0x5A) || (b >= 0x61 && b<= 0x7A))
                sb.append((char)b);
            else
                encode2184(sb, b);
        }
        return(sb.toString());
    }
    
    
    class FichierAttache {
        public String cid; // cid
        public String nom; // nom
        public InputStream contenu; // flux pour obtenir le contenu du fichier attaché
        public String mime; // type MIME
        
        public FichierAttache(final String cid, final String nom, final InputStream contenu, final String mime) {
            this.cid = cid;
            this.nom = nom;
            this.contenu = contenu;
            this.mime = (mime != null ? mime : "application/octet-stream");
        }
        
        public void envoyer(final BufferedWriter bw) throws IOException {
            bw.write("Content-Disposition: inline");
            if (nom != null) {
                envoyerLigne(";", bw);
                envoyerLigne("\tfilename*=" + encode2184(nom), bw);
            } else
                envoyerLigne("", bw);
            envoyerLigne("Content-Type: " + mime + ";", bw);
            envoyerLigne("\tname=\"" + encode1522(nom) + "\"", bw);
            if (cid != null)
                envoyerLigne("Content-Id: " + cid, bw);
            envoyerLigne("Content-Transfer-Encoding: base64", bw);
            envoyerLigne("", bw);
            final EncodeurBase64 encodeur = new EncodeurBase64(bw);
            encodeur.setFinDeLigne("\r\n");
            int length = 32;
            final byte[] ch = new byte[length];
            while ((length = contenu.read(ch)) != -1)
                encodeur.write(ch, 0, length);
            encodeur.fin();
            envoyerLigne("", bw);
        }
    }
    
}
