/*
PAELLOS - Plateforme d'Apprentissage En Ligne Libre Orientée Sémantique

Copyright (C) 2011 Observatoire de Paris

Ce programme est un logiciel libre ; vous pouvez le redistribuer et/ou le modifier conformément aux dispositions de la Licence Publique Générale GNU, telle que publiée par la Free Software Foundation ; version 2 de la licence, ou encore (à votre choix) toute version ultérieure.

Ce programme est distribué dans l'espoir qu'il sera utile, mais SANS AUCUNE GARANTIE ; sans même la garantie implicite de COMMERCIALISATION ou D'ADAPTATION A UN OBJET PARTICULIER. Pour plus de détail, voir la Licence Publique Générale GNU .

Vous devez avoir reçu un exemplaire de la Licence Publique Générale GNU en même temps que ce programme ; si ce n'est pas le cas, écrivez à la Free Software Foundation Inc., 675 Mass Ave, Cambridge, MA 02139, Etats-Unis.
*/

package paellos;

import java.io.*;
import java.util.Properties;
import javax.servlet.*;
import javax.servlet.http.*;

import org.apache.lucene.analysis.Analyzer;
import org.apache.lucene.analysis.standard.StandardAnalyzer;
import org.apache.lucene.document.Document;
import org.apache.lucene.index.IndexReader;
import org.apache.lucene.search.Hits;
import org.apache.lucene.search.IndexSearcher;
import org.apache.lucene.search.Query;
import org.apache.lucene.queryParser.ParseException;
import org.apache.lucene.queryParser.QueryParser;
import de.iqcomputing.lucene.BoldHighlighter;
import de.iqcomputing.lucene.LuceneTools;

import paellos.base.*;

/**
 * Recherche de pages sur le site.
 * Basé sur les JSP de Andrew C. Oliver, SuperLink Software, Inc. (acoliver2@users.sourceforge.net)
 * servant de démo pour Lucene.
 *
 * @author Damien Guillaume
 */
public class Recherche extends HttpServlet {
    
    String cheminSiteEtudiants;
    String cheminSiteTuteurs;
    String indexLucene; // chemin absolu vers le l'index Lucene sur le disque
    
    public void init(ServletConfig config) throws ServletException {
        super.init(config);
        final Properties props = new Properties();
        try {
            props.load(new InputStreamReader(getClass().getResourceAsStream("ServletProperties"), "UTF-8"));
            cheminSiteEtudiants = props.getProperty("cheminSiteEtudiants");
            cheminSiteTuteurs = props.getProperty("cheminSiteTuteurs");
            indexLucene = props.getProperty("indexLucene");
        } catch (IOException ex) {
            log(Strings.get("erreur_proprietes") + " " + ex.getMessage());
        }
    }
    
    /**
     * Recherche de page à partir d'une liste de mots-clés
     *
     * paramètres
     * - requete : liste de mots-clés
     * - debut : numéro de la page à partir de laquelle commencer l'affichage (entier)
     * - maxres : nombre maximum de résultats par page (entier)
     * 
     * Suppose URIEncoding="UTF-8" dans le Connector Tomcat
     */
    public void doGet(HttpServletRequest req, HttpServletResponse resp) throws ServletException, IOException {
        final HttpSession session = req.getSession(false);
        if (session == null) {
            log("Recherche.doGet: " + Strings.get("erreur_pas_de_session") + " " + req.getRequestURL());
            envoyerErreur(resp, Strings.get("session_expiree"));
            return;
        }
        final User user = (User)session.getAttribute("user");
        if (user == null) {
            log("Recherche.doGet: " + Strings.get("erreur_user_null") + " " + req.getRequestURL());
            envoyerErreur(resp, Strings.get("erreur_donnees_session"));
            return;
        }
        String requete = req.getParameter("requete");
        String sdebut = req.getParameter("debut");
        String smaxres = req.getParameter("maxres");
        if (requete == null) {
            afficherPageRecherche(resp);
            return;
        }
        int debut;
        if (sdebut == null)
            debut = 0;
        else {
            try {
                debut = (new Integer(sdebut)).intValue();
                if (debut < 0)
                    debut = 0;
            } catch (NumberFormatException ex) {
                envoyerErreur(resp, Strings.get("erreur_debut"));
                return;
            }
        }
        int maxres;
        if (smaxres == null)
            maxres = 50;
        else {
            try {
                maxres = (new Integer(smaxres)).intValue();
                if (maxres < 1)
                    maxres = 1;
            } catch (NumberFormatException ex) {
                envoyerErreur(resp, Strings.get("erreur_res_page_entier"));
                return;
            }
        }
        afficherResultats(resp, requete, debut, maxres, user);
    }
    
    public void afficherPageRecherche(final HttpServletResponse resp) {
        try {
            resp.setContentType("text/html; charset=UTF-8");
            final PrintWriter pw = resp.getWriter();
            afficherEntete(pw);
            pw.println("<div align=\"center\">");
            pw.println("<form name=\"recherche\" action=\"Recherche\" method=\"get\">");
            pw.println("<p>");
            pw.println("<input name=\"requete\" size=\"44\">");
            pw.println("</p>");
            pw.println("<p>");
            pw.println("<input name=\"maxres\" size=\"4\" value=\"50\">&nbsp;" + Strings.get("resultats_par_page") + "&nbsp;");
            pw.println("<input type=\"submit\" value=\"" + Strings.get("rechercher") + "\">");
            pw.println("</p>");
            pw.println("</form>");
            pw.println("</div>");
            afficherPiedDePage(pw);
            pw.flush();
        } catch (IOException ex) {
            log("afficherPageRecherche: IOException:" + ex.getMessage());
        }
    }
    
    public void afficherResultats(final HttpServletResponse resp, final String requete, final int debut, final int maxres, final User user) {
        
        try {
            resp.setContentType("text/html; charset=UTF-8");
            final PrintWriter pw = resp.getWriter();
            
            boolean error = false;                  // utilisé pour gérer le flux des messages d'erreur
            IndexSearcher searcher = null;          // utilisé pour ouvrir et rechercher dans l'index
            Query query = null;                     // la requête créée par le parser
            Hits hits = null;                       // les résultats de la requête
            int nbres = 0;                          // nb lignes renvoyées
            Analyzer analyseur = null;
            
            try {
                searcher = new IndexSearcher( IndexReader.open(indexLucene) );     // chercheur dans l'index
            } catch (IOException e) {
                // Erreur à l'ouverture de l'index: probablement un problème de permission, d'inexistance ou de corruption de l'index
                log("Recherche.afficherResultats: " + Strings.get("erreur_ouverture_index") + " " + e.getMessage());
                envoyerErreur(resp, Strings.get("erreur_interne_recherche"));
                return;
            }
            
            afficherEntete(pw);
            
            // analyseur syntaxique de requête avec des mots d'arrêt dans la bonne langue
            analyseur = new StandardAnalyzer(Strings.get("stop_words").split("\\s"));
            
            try {
                query = QueryParser.parse(requete, "contents", analyseur);
            } catch (ParseException e) {
                pw.println("<p>" + Strings.get("erreur_analyse_requete") + e.getMessage() + "</p>");
                error = true;
            }
            
            pw.println("<form name=\"recherche\" action=\"Recherche\" method=\"get\">");
            pw.println("<p><input name=\"maxres\" type=\"hidden\" value=\"" + maxres + "\">" +
                "<input name=\"requete\" size=\"44\" value=\"" + entites(requete) + "\"> " +
                "<input type=\"submit\" value=\"" + Strings.get("rechercher") + "\"></p>");
            pw.println("</form>");
            
            if (error == false && searcher != null) {                 // if we've had no errors
                                                                      // searcher != null was to handle a weird compilation bug 
                nbres = maxres;                                       // par défaut, on renvoie le maximum de lignes
                hits = searcher.search(query);                        // lancement de la requête
                if (hits.length() == 0) {                             // s'il n'y a pas de résultat on prévient l'utilisateur
                    pw.println("<p>" + Strings.get("rien_trouve") + "</p>");
                    error = true;                                     // et on s'arrête là
                }
            }
            
            if (error == false && searcher != null) {
                pw.println("<table>");
                pw.println("<tr>");
                pw.println("<th>" + Strings.get("document") + "</th>");
                pw.println("<th>" + Strings.get("extrait") + "</th>");
                pw.println("</tr>");
                if ((debut + maxres) > hits.length()) {
                    nbres = hits.length() - debut;
                }
                
                for (int i = debut; i < (nbres + debut); i++) {  // pour chaque résultat de la requête
                    pw.println("<tr>");
                    final Document doc = hits.doc(i);              // obtention du document
                    
                    String titredoc = doc.get("title");            // du titre
                    String cheminSite;
                    if ("étudiant".equals(user.fonction))
                        cheminSite = cheminSiteEtudiants;
                    else
                        cheminSite = cheminSiteTuteurs;
                    final String fichier = doc.get("fichier");
                    String url = cheminSite + fichier.substring(fichier.indexOf('/')+1);     // de l'URL
                    if ((titredoc == null) || titredoc.equals("")) // utilisation de l'URL s'il n'y a pas de titre
                        titredoc = url;
                    
                    //File f = new File(basesite + fichier.substring(fichier.indexOf('/')));
                    //String resume = LuceneTools.highlightTerms(f, new BoldHighlighter(), query, analyseur);
                    //if (resume.startsWith("java.io.FileNotFoundException"))
                    //    resume = "?";
                    final String contents = doc.get("contents");
                    String resume;
                    if (contents == null)
                        resume = "?";
                    else
                        resume = LuceneTools.highlightTerms(contents, new BoldHighlighter(), query, analyseur);
                    
                    int ntab;
                    if ((i - debut)/2 == (i - debut)/2.0)
                        ntab = 1;
                    else
                        ntab = 2;
                    
                    // affichage des résultats
                    pw.println("<td class=\"t" + ntab + "\"><a href=\"" + url + "\">" + titredoc + "</a></td>");
                    pw.println("<td class=\"t" + ntab + "\">" + resume + "</td>");
                    pw.println("</tr>");
                }
                
                if (debut > 0 || debut + maxres < hits.length()) {   // s'il y a d'autres résultats...
                    pw.println("<tr>\n<td></td><td>");
                    // contruction du lien "résultats précédents"
                    if (debut > 0) {
                        String urlprecedents = "Recherche?requete=" + requete + "&amp;maxres=" + maxres + "&amp;debut=" + (debut - maxres);
                        pw.print("<a href=\"" + urlprecedents + "\">" + "&lt;&lt; " + Strings.get("resultats_precedents") + "</a>");
                    }
                    // contruction du lien "résultats suivants"
                    if (debut + maxres < hits.length()) {
                        if (debut > 0)
                            pw.print(" &nbsp; ");
                        String urlsuivants = "Recherche?requete=" + requete + "&amp;maxres=" + maxres + "&amp;debut=" + (debut + maxres);
                        pw.print("<a href=\"" + urlsuivants + "\">" + Strings.get("resultats_suivants") + " &gt;&gt;</a>");
                    }
                    pw.println("</td>\n</tr>");
                }
                pw.println("</table>");
            }
            afficherPiedDePage(pw);
            pw.flush();
        } catch (IOException ex) {
            log("afficherResultats: IOException:" + ex.getMessage());
        }
    }
    
    public void afficherEntete(final PrintWriter pw) {
        pw.println("<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01 Transitional//EN\" \"http://www.w3.org/TR/html4/loose.dtd\">");
        pw.println("<html>");
        pw.println("<head>");
        pw.println("<meta http-equiv=\"Content-Type\" content=\"text/html; charset=UTF-8\">");
        pw.println("<title>" + Strings.get("recherche_pages") + "</title>");
        pw.println(" <link href=\"paellos.css\" rel=\"stylesheet\" type=\"text/css\">");
        pw.println(" <script type=\"text/javascript\" language=\"JavaScript\">");
        pw.println("  <!--");
        pw.println("  function selection() {");
        pw.println("   document.recherche.requete.focus();");
        pw.println("  }");
        pw.println("  //-->");
        pw.println(" </script>");
        pw.println("</head>");
        pw.println("<body onLoad=\"selection()\">");
        pw.println("<div align=\"center\">");
        pw.println("<img src=\"logo.png\" alt=\"logo\" align=\"right\">");
        pw.println("<h1>" + Strings.get("recherche_pages") + "</h1>");
        pw.println("</div>");
    }
    
    public void afficherPiedDePage(final PrintWriter pw) {
        pw.println("<p>" + Strings.get("recherche_avec_moteur") + " <a href=\"http://jakarta.apache.org/lucene/docs/index.html\">Lucene</a></p>");
        pw.println("</body>");
        pw.println("</html>");
    }
    
    protected void envoyerErreur(final HttpServletResponse resp, final String errmsg) {
        try {
            resp.setHeader("Pragma", "no-cache");
            resp.setHeader("Cache-Control", "no-cache");
            resp.setHeader("Expires", "0");
            resp.setContentType("text/html; charset=UTF-8");
            final PrintWriter pw = resp.getWriter();
            pw.println("<html>");
            pw.println("<head>");
            pw.println("<meta http-equiv=\"Content-Type\" content=\"text/html; charset=UTF-8\">");
            pw.println("<title>" + Strings.get("titre_erreur") + "</title>");
            pw.println("</head>");
            pw.println("<body>");
            pw.println(Strings.get("une_erreur") + " " + errmsg);
            pw.println("<br>");
            pw.println(Strings.get("bouton_retour"));
            pw.println("</body>");
            pw.println("</html>");
            pw.flush();
        } catch (IOException ex) {
            log("envoyerErreur: " + ex.getMessage());
        }
    }
    
    // encodage d'entités HTML pour l'encodage UTF-8
    public static String entites(final String s) {
        if (s == null)
            return("");
        String res = "";
        for (int i=0; i<s.length(); i++) {
            char c = s.charAt(i);
            if (c == '"')
                res += "&quot;";
            else if (c == '&')
                res += "&amp;";
            else if (c == '<')
                res += "&lt;";
            else if (c == '>')
                res += "&gt;";
            else
                res += c;
        }
        res = res.replaceAll("\\p{Cc}", "?"); // caractères de contrôle
        return(res);
    }
}
